import napalm.package
import napalm_core


def entry_point():
    """This is the entry point for the napalm package.

    It provisions your detection modules and provides them to napalm.

    It returns a dictionary of Collections, keyed by the name of the collection.
    """
    return napalm.package.entry_point(napalm_core)



from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.cfg.node import NodeType
from slither.core.expressions.member_access import MemberAccess
from slither.core.expressions.call_expression import CallExpression
from toolz.curried import pipe, map, filter, reduce
from operator import add


class DoSPushArrayLoop(AbstractDetector):
    ARGUMENT = "dos-push-array-loop"
    HELP = "Potential Denial of Service when users can push to an array that's irreducible and looped over."
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = (
        DetectorClassification.MEDIUM
    )  # it's impossible to determine whether a fee is actually used

    WIKI = "None"

    WIKI_TITLE = "Denial of Service on boundless loops over arrays"
    WIKI_DESCRIPTION = "Detects loops over arrays that users can push to, but that can't be popped from"
    WIKI_EXPLOIT_SCENARIO = "An attacker would push elements into the array until it becomes impossible to loop over."
    WIKI_RECOMMENDATION = (
        """Refactor the code to not loop over user controlled arrays."""
    )

    # NOTE: slither doesn't provide a post-domination tree, so we're unable to filter false positives when a pop
    # always follows each push. The following fp prompt is a remediation.
    FALSE_POSITIVE_PROMPT = """
        This is a false positive if the contract pops from the array after pushing to it
        This is a false positive if the contract resets the length of the array.
    """

    def _array_pushes_in_public_functions(self):
        return pipe(
            self.compilation_unit.functions,
            filter(lambda f: f.visibility == "public"),
            map(lambda f: f.nodes),
            reduce(add),
            filter(lambda n: n.type == NodeType.EXPRESSION),
            filter(lambda n: n.expression is not None),
            filter(lambda n: isinstance(n.expression, CallExpression)),
            filter(lambda n: isinstance(n.expression.called, MemberAccess)),
            filter(lambda n: n.expression.called.member_name == "push"),
        )

    def _array_pops_in_any_functions(self):
        # this would improve precision of the module
        pass

    def _loops_over_arrays(self):
        return pipe(
            self.compilation_unit.functions,
            map(lambda f: f.nodes),
            reduce(add),
            filter(lambda n: n.type == NodeType.IFLOOP),
        )

    def _checks(self, pushes, loops):
        tainted_storage_variables = pipe(
            pushes, map(lambda n: (n.state_variables_written, n)), list
        )

        for loop in loops:
            if loop.expression is None:
                continue
            if loop.state_variables_read is None:
                continue

            result = self.generate_result([loop, self.HELP])

            write_locations = []
            for variable in loop.state_variables_read:
                for variables, location in tainted_storage_variables:
                    if variable in variables:
                        write_locations.append(location)
                        result.add_variable(variable)
                        result.add_node(location)

            if not write_locations:
                continue

            yield result

    def _detect(self):
        pushes = list(self._array_pushes_in_public_functions())
        loops = list(self._loops_over_arrays())

        return list(self._checks(pushes, loops))



import re
from typing import List, Tuple

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
)
from slither.utils.output import Output

PATTERN = re.compile(r"(\^|>|>=|<|<=)?([ ]+)?(\d+)\.(\d+)\.(\d+)")


class Push0SolidityVersionDetector(AbstractDetector):
    ARGUMENT = "push-0-solc-version"
    HELP = "Your solidity version uses PUSH0 which is not supported by all chains."
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.LOW
    LANGUAGE = "solidity"
    WIKI = "None"

    WIKI_TITLE = "Push0 Solidity Version Detector"

    WIKI_DESCRIPTION = (
        """Your solidity version uses PUSH0 which is not supported by all chains."""
    )
    WIKI_RECOMMENDATION = """ Deploy with a version < 0.8.20"""

    def _check_version(self, version: Tuple[str, str, str, str, str]) -> bool:
        op, _, major, minor, inter = version
        if op and op in [">", ">=", "^"]:
            return True
        elif op == "<" and int(major) >= 8 and int(inter) <= 20:
            return False
        elif op == "<=" and int(major) >= 8 and int(inter) < 20:
            return False
        else:
            return True

    def _check_pragma(self, version: str) -> bool:
        versions = PATTERN.findall(version)
        if len(versions) == 1:
            version = versions[0]
            return self._check_version(version)
        elif len(versions) == 2:
            return self._check_version(versions[0]) or self._check_version(versions[1])

    def _detect(self) -> List[Output]:
        results = []
        pragma = self.compilation_unit.pragma_directives

        for p in pragma:
            if len(p.directive) < 1 or p.directive[0] != "solidity":
                continue

            # This is version, so we test if this is disallowed.
            is_above_0_18_19 = self._check_pragma(p.version)

            if is_above_0_18_19:
                results.append(self.generate_result([p, self.HELP]))

        return results



from typing import Set, List, TYPE_CHECKING

from slither.core.cfg.node import NodeType

from slither.core.cfg.node import Node

from toolz import reduce


"""
Algorithm and implementation of post-dominators is largely based on original dominator implementation in slither.

Source: https://github.com/crytic/slither/blob/46be017fe538c04dbc5928c1134d78e8c3e70e74/slither/core/dominators/utils.py
"""

from copy import copy


def get_post_dominators(node: "Node") -> Set["Node"]:
    if hasattr(node, "post_dominators"):
        # return copy(node.post_dominators)
        return node.post_dominators
    else:
        return set()


def set_post_dominators(node: Node, nodes: List["Node"]) -> None:
    node.post_dominators = nodes


def set_immediate_post_dominator(node: Node, immediate_post_dominator: "Node") -> None:
    node.immediate_post_dominator = immediate_post_dominator


def get_immediate_post_dominator(node: Node) -> "Node":
    if hasattr(node, "immediate_post_dominator"):
        return node.immediate_post_dominator
    else:
        return None


def _intersection_successor(node: "Node") -> Set["Node"]:
    if not node.sons:
        return set()

    if not any(son.is_reachable for son in node.sons):
        return set()

    return set(reduce(set.intersection, map(get_post_dominators, node.sons)))


def _compute_post_dominators(nodes: List["Node"]) -> None:
    changed = True
    while changed:
        changed = False
        for node in nodes:
            new_set = _intersection_successor(node).union({node})
            if new_set != node.post_dominators:
                node.post_dominators = new_set
                changed = True


def _compute_immediate_post_dominators(nodes: List["Node"]) -> None:
    for node in nodes:
        idom_candidates = get_post_dominators(node) - {node}

        if len(idom_candidates) == 1:
            set_immediate_post_dominator(node, idom_candidates.pop())
            continue

        all_dominators = set()
        for d in idom_candidates:
            if d in all_dominators:
                continue
            all_dominators |= get_post_dominators(d) - {d}

        idom_candidates = all_dominators.symmetric_difference(idom_candidates)
        assert len(idom_candidates) <= 1
        if idom_candidates:
            idom = idom_candidates.pop()
            set_immediate_post_dominator(node, idom)


def compute_post_dominators(nodes: List["Node"]) -> None:
    if not nodes:
        return

    for n in nodes:
        n.post_dominators = set(nodes)

    _compute_post_dominators(nodes)

    _compute_immediate_post_dominators(nodes)

    return nodes



from slither.core.declarations.function_contract import FunctionContract
from slither.detectors.abstract_detector import AbstractDetector
from typing import Callable
from slither.core.variables.state_variable import StateVariable
from slither.slithir.variables import TupleVariable
from slither.slithir.operations import Condition
from slither.slithir.operations import SolidityCall
from slither.slithir.operations import Assignment
from slither.core.cfg.node import NodeType, Function, Node

from toolz.curried import pipe, map, filter, reduce, count, concat, itemmap

from napalm_core.detectors.util.taint import taint_analysis, get_taint, set_taint
from slither.analyses.data_dependency.data_dependency import (
    is_dependent_ssa,
    is_dependent,
)

from uuid import uuid4


def detect_unused_return_values(tainted_variables: Callable, function: Function):
    _taint_identifier = f"unused-return-values{uuid4()}"

    def taint_function(node: Node):
        writes = list(tainted_variables(node))
        taint = [(node, writes)] if writes else []
        taint = sanitize_function(node, taint)
        set_taint(_taint_identifier, node, taint)

    def sanitize_function(node: Node, taint):
        def clear_taint_for(variables, _taint):
            for _tainted_node, _tainted_variables in _taint:
                result = _tainted_node, [
                    tainted_variable
                    for tainted_variable in _tainted_variables
                    if not any(
                        is_dependent(read_variable, tainted_variable, node.function)
                        for read_variable in variables
                    )
                ]
                if result[1]:
                    yield result

        def sanitize_if(_taint):
            _taint = _taint[:] if _taint else []
            if node.type != NodeType.IF:
                return _taint
            return list(clear_taint_for(node.variables_read, _taint))

        def sanitize_require(_taint):
            _taint = _taint[:] if _taint else []
            try:
                is_requires_check = (
                    node.type == NodeType.EXPRESSION
                    and node.expression.called.value.full_name == "require(bool)"
                )
            except AttributeError:
                is_requires_check = False
            if not is_requires_check:
                return _taint

            return list(clear_taint_for(node.variables_read, _taint))

        def sanitize_state_variable(_taint):
            _taint = _taint[:] if _taint else []
            if not node.state_variables_written:
                return _taint
            return list(clear_taint_for(node.state_variables_written, _taint))

        sanitized = pipe(
            taint,
            sanitize_if,
            sanitize_require,
            sanitize_state_variable,
        )

        return sanitized

    def sink_function(node: Node):
        return node.type != NodeType.THROW and len(node.sons) == 0

    def join_taint_function(base, other):
        base = base or []
        # base_names = [variable.name for (node, variables) in base for variable in variables]
        other = other or []
        other = [e for e in other if e not in base]
        return base + other

    return taint_analysis(
        taint_identifier=_taint_identifier,
        taint_function=taint_function,
        sanitise_function=sanitize_function,
        sink_function=sink_function,
        join_taint_function=join_taint_function,
        function_under_test=function,
    )



from slither.core.declarations import Function
from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.operations.unused_return_values import UnusedReturnValues
from slither.slithir.operations import HighLevelCall
from slither.slithir.operations.operation import Operation


class UncheckedApprove(UnusedReturnValues):
    ARGUMENT = "unchecked-approve"
    HELP = "Unchecked token approve"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "None"

    WIKI_TITLE = "Unchecked approve"
    WIKI_DESCRIPTION = "The return value of an external approve call is not checked"

    WIKI_EXPLOIT_SCENARIO = "None"

    WIKI_RECOMMENDATION = (
        "Use `SafeERC20`, or ensure that the approve return value is checked."
    )

    def _is_instance(self, ir: Operation) -> bool:
        return (
            isinstance(ir, HighLevelCall)
            and isinstance(ir.function, Function)
            and ir.function.solidity_signature in ["approve(address,uint256)"]
        )



from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import SolidityCall
from slither.core.declarations.solidity_variables import SolidityFunction
from slither.slithir.operations.operation import Operation
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.core.expressions.binary_operation import (
    BinaryOperation,
    BinaryOperationType,
)
from slither.slithir.operations.binary import Binary, BinaryType

from napalm_core.detectors.util.unused_return_values import detect_unused_return_values


class UncheckedEquationModifierDetector(AbstractDetector):
    ARGUMENT = "unchecked-equation-modifier"
    HELP = "Potential lack of input control due to missing requires checks."
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = (
        DetectorClassification.HIGH
    )  # it's impossible to determine whether a fee is actually used

    WIKI = "None"

    WIKI_TITLE = "Potential lack of input control due to missing requires checks."
    WIKI_DESCRIPTION = (
        "Detects strict equations in modifiers that are not checked / used."
    )
    WIKI_EXPLOIT_SCENARIO = "Undefined"
    WIKI_RECOMMENDATION = (
        """Ensure all checks are performed, for example, using require."""
    )


    @staticmethod
    def _tainted_variables(node):
        for ir in node.irs:
            if isinstance(ir, Binary) and ir.type == BinaryType.EQUAL:
                return node.variables_written
        return []

    def _scan(self):
        for function in self.compilation_unit.modifiers:
            for node in function.nodes:
                if (
                    isinstance(node.expression, BinaryOperation)
                    and node.expression.type == BinaryOperationType.EQUAL
                    and not node.variables_written
                ):
                    yield self.generate_result([node, self.HELP])

            result = detect_unused_return_values(self._tainted_variables, function)
            for tainted, node, variables in result:
                if variables:
                    yield self.generate_result([node, self.HELP])

    def _detect(self):
        return list(self._scan())



from slither.core.declarations.solidity_variables import SolidityFunction
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.slithir.operations import SolidityCall
from slither.slithir.operations.operation import Operation
from slither.core.cfg.node import Node

from napalm_core.detectors.util.unused_return_values import detect_unused_return_values


class UncheckedReturnCreate2(AbstractDetector):
    ARGUMENT = "unchecked-create-2"
    HELP = "Unchecked return value for CREATE2 call."
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "None"

    WIKI_TITLE = "Unchecked return value create 2"
    WIKI_DESCRIPTION = "The return value of a create 2 call is not checked"

    WIKI_EXPLOIT_SCENARIO = "None"

    WIKI_RECOMMENDATION = "Ensure that the return value of CREATE2 is checked."

    @staticmethod
    def tainted_variables(node: Node):
        for ir in node.irs:
            if (
                isinstance(ir, SolidityCall)
                and isinstance(ir.function, SolidityFunction)
                and ir.function.full_name == "create2(uint256,uint256,uint256,uint256)"
            ):
                return node.variables_written
        return []

    def _scan(self):
        for function in self.compilation_unit.functions:
            result = detect_unused_return_values(self.tainted_variables, function)
            for tainted, node, variables in result:
                yield self.generate_result([node, self.HELP])

    def _detect(self):
        return list(self._scan())



from operator import add

from slither.core.cfg.node import NodeType
from slither.core.expressions.call_expression import CallExpression
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from toolz.curried import pipe, map, filter, reduce


class ERC20ApproveCall(AbstractDetector):
    ARGUMENT = "erc20-approve-call"
    HELP = "Potential Denial of Service when tokens don't implement approve() consistently ( e.g. USDT )."
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.LOW

    WIKI = "None"

    WIKI_TITLE = "Denial of Service on non-standard ERC20 tokens"
    WIKI_DESCRIPTION = "Detects unsafe usage of approve(), which isn't consistently implemented amongst popular tokens."
    WIKI_EXPLOIT_SCENARIO = "Undefined"
    WIKI_RECOMMENDATION = (
        """Use the safeERC20 library which handles most of the edge cases."""
    )

    @staticmethod
    def _is_erc20(contract):
        for _contract in contract.inheritance:
            if "ERC20" in _contract.name or "EIP20" in _contract.name:
                return True
        if "ERC20" in contract.name or "EIP20" in contract.name:
            return True
        return False

    def _approve_calls(self):
        return pipe(
            self.compilation_unit.functions,
            map(lambda f: f.nodes),
            reduce(add),
            filter(lambda n: n.type == NodeType.EXPRESSION),
            filter(lambda n: n.expression is not None),
            filter(lambda n: isinstance(n.expression, CallExpression)),
            filter(lambda n: n.high_level_calls is not None),
            filter(
                lambda n: filter(
                    lambda c, f: self._is_erc20(c) and f.name == "approve",
                    n.high_level_calls,
                )
            ),
        )

    def _checks(self):
        for call in self._approve_calls():
            yield self.generate_result([call, self.HELP])

    def _detect(self):
        return list(self._checks())



from operator import add

from slither.core.cfg.node import NodeType
from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from toolz.curried import pipe, map, filter, reduce


class ERC721MintDetector(AbstractDetector):
    ARGUMENT = "erc721-mint-stuck"
    HELP = "_mint call on ERC721 can lead to stuck tokens"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "None"

    WIKI_TITLE = "ERC721 _mint Calls"
    WIKI_DESCRIPTION = "Detects calls to _mint on ERC721 tokens."
    WIKI_EXPLOIT_SCENARIO = 'Sample'
    WIKI_RECOMMENDATION = """Use `_safeMint` instead of `_mint` which ensures that receiving contracts
can handle receiving ERC721 tokens."""

    def _scan(self):
        erc721_contracts = []
        for contract in self.compilation_unit.contracts_derived:
            if "ERC721" in contract.name:
                erc721_contracts.append(contract.id)
                continue
            for parent in contract.inheritance:
                if "ERC721" in parent.name:
                    erc721_contracts.append(contract.id)

        for expression, high_level_calls in pipe(
            self.compilation_unit.contracts_derived,
            map(lambda c: c.functions),  # [[functions]]
            reduce(add),
            map(lambda f: f.nodes),  # [[nodes]]
            reduce(add),
            map(lambda e: (e, e.high_level_calls)),  # [[expression, high_level_calls]]
        ):
            for contract, function in high_level_calls:
                if contract.id in erc721_contracts and function.name == "_mint":
                    yield self.generate_result([expression, self.HELP])

    def _detect(self):
        return list(self._scan())



import copy
from typing import Callable, TypeVar, Generic, Optional

from slither.core.cfg.node import NodeType, Function

from slither.core.cfg.node import Node

from toolz import reduce, iterate, curry


@curry
def get_taint(taint_identifier: str, node: Node):
    return node.context.get(f"TAINT_{taint_identifier}", None)


def set_taint(taint_identifier: str, node: Node, taint):
    node.context[f"TAINT_{taint_identifier}"] = taint


def _spread_taint(
    taint_identifier: str, join_taint_function, sanitize_function, node: Node
):
    node_taint_before = get_taint(taint_identifier, node)

    new_taint = reduce(
        join_taint_function,
        map(get_taint(taint_identifier), node.fathers),
        copy.copy(node_taint_before),
    )

    new_taint = sanitize_function(node, new_taint)

    set_taint(taint_identifier, node, new_taint)

    updated = new_taint != node_taint_before

    return updated


def _is_tainted(taint_identifier: str, node: Node):
    return node.context.get(f"TAINT_{taint_identifier}", None) is not None


TaintType = TypeVar("TaintType")


# this says taint analysis, but it's really just a generic fixed point algorithm
def taint_analysis(
    taint_identifier: str,
    taint_function: Callable[[Node], None],
    sanitise_function: Callable[[Node], bool],
    sink_function: Callable[[Node], bool],
    join_taint_function: Callable[
        [Optional[TaintType], Optional[TaintType]], Optional[TaintType]
    ],
    function_under_test: Function,
):
    """Perform intra procedural taint analysis on a function.

    Args:
        taint_identifier: The identifier to use for the taint.
        taint_function: A function that takes a node and taints it.
        sanitise_function: A function that takes a node and returns whether it sanitises taint.
        sink_function: A function that takes a node and returns whether it is a sink.
        join_taint_function: A function that takes two taints and joins them.
        function_under_test: The function to perform the analysis on.

    Returns:
        Whether the function is tainted.
    """
    # taint
    for node in function_under_test.nodes:
        taint_function(node)

    # spread
    changed = True
    while changed:
        changed = False
        for node in function_under_test.nodes:
            changed |= _spread_taint(
                taint_identifier, join_taint_function, sanitise_function, node
            )

    # check
    for node in function_under_test.nodes:
        if sink_function(node) and _is_tainted(taint_identifier, node):
            yield True, node, get_taint(taint_identifier, node)



from typing import List

from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification
from slither.utils.type_helpers import HighLevelCallType
from toolz.curried import pipe, filter, count


class ChainlinkCircuitBreakerDetector(AbstractDetector):
    ARGUMENT = "chainlink-circuit-breaker"
    HELP = "Vulnerable to exploitation in case of a Chainlink circuit breaker event."
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = "None"

    WIKI_TITLE = "Chainlink Circuit Breaker"
    WIKI_DESCRIPTION = """
    Detects unsafe usage of chainlink oracles, where the contract might be vulnerable in case of a circuit breaker event.
    """

    WIKI_EXPLOIT_SCENARIO = "Undefined"
    WIKI_RECOMMENDATION = """Check the result of the chainlink oracle call to ensure it is within the expected range."""

    def _calls(self, function) -> List[HighLevelCallType]:
        for node in function.nodes:
            for high_level_call in node.high_level_calls:
                yield node, high_level_call

    def _checks(self):
        for function in self.compilation_unit.functions:
            _calls = list(self._calls(function))

            # exit(0)
            latest_round_calls = pipe(
                _calls,
                filter(
                    lambda _high_level_call: _high_level_call[1][1].name
                    == "latestRoundData"
                ),
                list,
            )
            min_value = pipe(
                _calls,
                filter(
                    lambda _high_level_call: _high_level_call[1][1].name == "minValue"
                ),
                count,
            )
            max_value = pipe(
                _calls,
                filter(
                    lambda _high_level_call: _high_level_call[1][1].name == "maxValue"
                ),
                count,
            )

            if not latest_round_calls:
                continue
            if min_value and max_value:
                # Improve: check if min_value and max_value are used
                # tbh: if both are queried then it's likely that they're used in a check
                continue

            for call in latest_round_calls:
                yield self.generate_result([call[0], self.HELP])

    def _detect(self):
        return list(self._checks())
